@@ -6,25 +6,22 @@
 
 """Test bidict metadata."""
 
+from __future__ import annotations
+
 import bidict
 
 
 METADATA_ATTRS = """
 __author__
-__maintainer__
 __copyright__
-__email__
 __description__
-__keywords__
 __license__
-__status__
 __url__
 __version__
-__project_urls__
 """.split()
 
 
-def test_metadata():
+def test_metadata() -> None:
     """Ensure bidict has expected metadata attributes."""
     for i in METADATA_ATTRS:
         assert getattr(bidict, i)
